//: /be/optiloading/gui/MainFrame.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2007 Tim Pintens. All rights reserved.
 */

package be.optiloading.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Toolkit;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.ResourceBundle;
import java.util.prefs.Preferences;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

import be.optiloading.cargo.CargoPanel;
import be.optiloading.compatibility.CargoCompPanel;
import be.optiloading.compatibility.TankCompPanel;
import be.optiloading.generalinfo.InfoPanel;
import be.optiloading.helpers.files.Doc;
import be.optiloading.settings.Settings;
import be.optiloading.ship.IntactStability;
import be.optiloading.ship.ShipData;
import be.optiloading.ship.StabilityGraph;
import be.optiloading.ship.StrengthSummary;
import be.optiloading.tank.TankPanel;

/**
 * MainFrame class creates the general frame.
 * 
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank"
 *         href="http://www.optiloading.be">www.optiloading.be</a>
 */

@SuppressWarnings("serial")
public class MainFrame extends JFrame{
	private JTabbedPane tabbedPane;
	private TankPanel ballastTanks;
	private TankPanel cargoTanks;
	private InfoPanel infoPanel;
	private TankPanel miscTanks;
	private CargoPanel cargoPanel;
	private CargoCompPanel cargoCompPanel;
	private TankCompPanel tankCompPanel;
	private StabilityGraph stabilityGraph;
	private IntactStability intactStability;
	private StrengthSummary strengthSummary;
	private Preferences prefs;
	
	private int mainsplit;
	private int compatibilitysplit;
	private int inputsplit;
	private int outputsplit;
	private int stabilitysplit;
	private int mainheight;
	private int mainwidth;
	private int mainx;
	private int mainy;
	private int extendedstate;
	
	private JPanel cards;
	private JSplitPane splitPane;
	private JSplitPane inputpane;
	private JSplitPane outputpane;
	private JSplitPane stabilitypane;
	private JSplitPane compatibilitypane;
	
	private JPanel inputContainer;
	private JPanel compatibilityContainer;
	private JPanel outputContainer;
	private JPanel stabilityContainer;
	
	public final static String MAINPANEL = "Main Window";
    public final static String PROGRESSPANEL = "Progress";
	
	private static MainFrame uniqueInstance;
	
	private ResourceBundle languageResource;
	
	private MainMenu mainMenu;
	
	private boolean conventional;
	
	/**
	 * Returns the uniqueInstance of the MainFrame object.
	 * 
	 * @return MainFrame object.
	 */
	public static MainFrame getInstance() {
		if (uniqueInstance == null) {
			uniqueInstance = new MainFrame();
		}
		return uniqueInstance;
	}

	//Singleton constructor
	private MainFrame(){
		super();
		prefs = Preferences.userNodeForPackage(getClass());
		languageResource = Settings.getInstance().getResourceBundle();
		
		Toolkit toolkit = Toolkit.getDefaultToolkit();
		toolkit.getScreenSize();
		
		loadPrefs();
		
		tabbedPane = new JTabbedPane();
		
		if(conventional){
			toConventional(true);
		} else {
			toAutomatic(true);
		}
		
		infoPanel = InfoPanel.getInstance();
		
		splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
				infoPanel,tabbedPane);
		splitPane.setOneTouchExpandable(true);
		splitPane.setDividerLocation(mainsplit);
		splitPane.addPropertyChangeListener(JSplitPane.DIVIDER_LOCATION_PROPERTY, 
				new PropertyChangeListener(){
			public void propertyChange(PropertyChangeEvent arg0) {
				if(splitPane.getLastDividerLocation() == -1){
					splitPane.setDividerLocation(270);
				}
			}
		});

		cards = new JPanel(new CardLayout());
		cards.add(splitPane, MAINPANEL);
		cards.add(ProgressPanel.getInstance(), PROGRESSPANEL);
		
		getContentPane().add(BorderLayout.CENTER, cards);
        
		setIconImage(new ImageIcon(this.getClass().getResource(
		"/images/icon.png")).getImage());
		mainMenu = new MainMenu(conventional);
		setJMenuBar(mainMenu);
		
		setSize(mainwidth, mainheight);
		setLocation(mainx, mainy);
		setExtendedState(extendedstate);
		setAlwaysOnTop(true);
		setTitle(languageResource.getString("programTitle")+" - "+
				languageResource.getString("untitled"));
		
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		addWindowListener(new WindowAdapter(){
			public void windowClosing(WindowEvent arg0) {
				saveAndClose();
			}
		});
		
		addComponentListener(new ComponentAdapter(){
			public void componentResized(ComponentEvent arg0) {
				outputpane.setDividerLocation((Integer) outputpane.getSize().height/2);
				stabilitypane.setDividerLocation((Integer) stabilitypane.getSize().height-200);
				if(!conventional){
					inputpane.setDividerLocation((Integer) inputpane.getSize().height/2);
					compatibilitypane.setDividerLocation((Integer) inputpane.getSize().height/2);
				}
			}
		});
	}
	 // Load preferences
	private void loadPrefs(){
		mainsplit = prefs.getInt("mainsplit", 270);
		compatibilitysplit = prefs.getInt("compatibilitysplit", 0);
		inputsplit = prefs.getInt("inputsplit", 0);
		outputsplit = prefs.getInt("outputsplit", 0);
		stabilitysplit = prefs.getInt("stabilitysplit", 600);
		mainheight = prefs.getInt("mainheight", 800);
		mainwidth = prefs.getInt("mainwidth", 1200);
		mainx = prefs.getInt("mainx", 0);
		mainy = prefs.getInt("mainy", 0);
		conventional = prefs.getBoolean("conventional", false);
		extendedstate = prefs.getInt("extendedstate", JFrame.MAXIMIZED_BOTH);
	}
	
	/**
	 * Method to quit the program.
	 */
	public void saveAndClose(){
		if(Doc.getInstance().isCanged()){
			int returnvalue = JOptionPane.showConfirmDialog(this,
					languageResource.getString("saveFile"), 
					languageResource.getString("saveChanges"),
			        JOptionPane.YES_NO_CANCEL_OPTION,
			        JOptionPane.QUESTION_MESSAGE);
			if (returnvalue == JOptionPane.YES_OPTION){
				if(mainMenu.save()){
					closeProgram();
				}
			} else if(returnvalue == JOptionPane.NO_OPTION){
				closeProgram();
			}
		} else {
			closeProgram();
		}
		
	}
	
	// Method to save the settings and close the program
	private void closeProgram(){
		prefs.putInt("mainsplit", splitPane.getDividerLocation());
		if(!conventional){
			prefs.putInt("inputsplit", inputpane.getDividerLocation());
			prefs.putInt("compatibilitysplit", compatibilitypane.getDividerLocation());
		}
		prefs.putInt("outputsplit", outputpane.getDividerLocation());
		prefs.putInt("stabilitysplit", stabilitypane.getDividerLocation());
		prefs.putInt("mainheight", getSize().height);
		prefs.putInt("mainwidth", getSize().width);
		prefs.putInt("mainx", getLocation().x);
		prefs.putInt("mainy", getLocation().y);
		prefs.putInt("extendedstate", getExtendedState());
		prefs.putBoolean("conventional", conventional);
		System.exit(0);
	}
	
	// Create tabs for conventional mode
	private void createConvTabs(){
		ShipData data = ShipData.getInstance();
		
		//Input
		miscTanks =  new TankPanel(languageResource.getString("miscCompartments"),
				true,false,false, data.getMiscTankList());
		cargoPanel = new CargoPanel();
		inputpane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
				cargoPanel,miscTanks);
		inputpane.setDividerLocation(inputsplit);
		tabbedPane.addTab(languageResource.getString("miscCompartments"), 
				new ImageIcon(this.getClass().getResource("/images/table.png")), 
				miscTanks,languageResource.getString("miscCompartmentsTooltip"));
		
		//Output
		ballastTanks = new TankPanel(languageResource.getString("ballastTanks")
				,true,false,true, data.getBallastList());
		
		cargoTanks = new TankPanel(languageResource.getString("cargoTanks"),
				true, true,false,data.getCargoTankList());

		outputpane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
				cargoTanks, ballastTanks);
		outputContainer = new JPanel(new BorderLayout());
		outputContainer.add(BorderLayout.CENTER,outputpane);
		outputpane.setDividerLocation(outputsplit);
		tabbedPane.addTab(languageResource.getString("weightDist"), 
				new ImageIcon(this.getClass().getResource("/images/table.png")), 
				outputContainer,languageResource.getString("weightDistTooltip"));
	}
	
	// Create tabs for automatic mode
	private void createAutTabs(){
		ShipData data = ShipData.getInstance();
		
		//Input
		miscTanks =  new TankPanel(languageResource.getString("miscCompartments"),
				true,false,false, data.getMiscTankList());
		cargoPanel = new CargoPanel();
		inputpane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
				cargoPanel,miscTanks);
		inputpane.setDividerLocation(inputsplit);
		inputContainer = new JPanel(new BorderLayout());
		inputContainer.add(BorderLayout.CENTER,inputpane);
		tabbedPane.addTab(languageResource.getString("input"), 
				new ImageIcon(this.getClass().getResource("/images/loadlist.gif")), 
				inputContainer,languageResource.getString("miscCompartmentsTooltip"));
		
		//Compatibility
		cargoCompPanel = new CargoCompPanel();
		tankCompPanel = new TankCompPanel();
		compatibilitypane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, 
				cargoCompPanel, tankCompPanel);
		compatibilitypane.setDividerLocation(compatibilitysplit);
		compatibilityContainer = new JPanel(new BorderLayout());
		compatibilityContainer.add(BorderLayout.CENTER,compatibilitypane);
		tabbedPane.addTab(languageResource.getString("compatibility"), 
				new ImageIcon(this.getClass().getResource("/images/danger.png")), 
				compatibilityContainer,languageResource.getString("miscCompatibilityTooltip"));
		
		//Output
		ballastTanks = new TankPanel(languageResource.getString("ballastTanks")
				,false,false,true, data.getBallastList());
		cargoTanks = new TankPanel(languageResource.getString("cargoTanks"),
				false, true,false,data.getCargoTankList());
		outputpane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
				cargoTanks, ballastTanks);
		outputContainer = new JPanel(new BorderLayout());
		outputContainer.add(BorderLayout.CENTER,outputpane);
		outputpane.setDividerLocation(outputsplit);
		tabbedPane.addTab(languageResource.getString("weightDist"), 
				new ImageIcon(this.getClass().getResource("/images/table.png")), 
				outputContainer,languageResource.getString("weightDistTooltip"));
	}
	
	// Create tabs that are always needed
	private void createTabs(){
		//Stability
		stabilityGraph = new StabilityGraph();
		intactStability = new IntactStability();
		JPanel holdgraph = new JPanel(new BorderLayout());
		holdgraph.setBorder(new CompoundBorder(new TitledBorder(new EtchedBorder(),
		"GZ curve"), new EmptyBorder(5, 5, 0, 5)));
		holdgraph.add(BorderLayout.CENTER,stabilityGraph);
		
		stabilitypane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
				holdgraph, intactStability);
		stabilitypane.setDividerLocation(stabilitysplit);
		stabilityContainer = new JPanel(new BorderLayout());
		stabilityContainer.add(BorderLayout.CENTER,stabilitypane);
		tabbedPane.addTab(languageResource.getString("stability"), 
				new ImageIcon(this.getClass().getResource("/images/stab.png")), 
				stabilityContainer,languageResource.getString("stabilityTootip"));
		
		//Strength
		strengthSummary = new StrengthSummary();
		tabbedPane.addTab(languageResource.getString("strength"), 
				new ImageIcon(this.getClass().getResource("/images/chart.png")), 
				strengthSummary,languageResource.getString("strengthTootip"));
	}
	
	/**
	 * Update GUI components.
	 */
	public void updateInterface(){
		if(!conventional){
			cargoCompPanel.update();
			tankCompPanel.update();
			cargoPanel.update();
		}
		ballastTanks.update();
		cargoTanks.update();
		miscTanks.update();
		infoPanel.update();
		intactStability.update();
		strengthSummary.update();
		stabilityGraph.repaint();
		Doc.getInstance().update();
		mainMenu.update();
	}
	
	/**
	 * Method to show specific panel. Either progresspanel or main interface.
	 * @param panel String containing name of panel.
	 */
	public void showPanel(String panel){
		CardLayout cl = (CardLayout)(cards.getLayout());
        cl.show(cards, panel);
	}
	
	/**
	 * Lock GUI components.
	 */
	public void lock(){
		mainMenu.lock();
		showPanel(PROGRESSPANEL);
	}
	
	/**
	 * Unlock GUI components.
	 */
	public void unlock(){
		mainMenu.unlock();
		showPanel(MAINPANEL);
		ProgressPanel.getInstance().stopCounter();
	}
	
	/**
	 * Make a specific tab active.
	 * @param tab Integer containing which tab to activate.
	 */
	public void setTabbedPane(int tab){
		tabbedPane.setSelectedIndex(tab);
	}
	
	/**
	 * Switch to conventional mode.
	 * @param startup Indicates if the application is starting up.
	 */
	public void toConventional(boolean startup){
		conventional = true;
		int currenttab = tabbedPane.getSelectedIndex();
		if(currenttab > 1){
			currenttab--;
		}
		tabbedPane.removeAll();
		createConvTabs();
		createTabs();
		tabbedPane.revalidate();
		ShipData.getInstance().setConventional(true);
		if(!startup){
			setTabbedPane(currenttab);
		}
	}
	
	/**
	 * Switch to automatic mode.
	 * @param startup Indicates if the application is starting up.
	 */
	public void toAutomatic(boolean startup){
		conventional = false;
		int currenttab = tabbedPane.getSelectedIndex();
		if(currenttab > 0){
			currenttab++;
		}
		tabbedPane.removeAll();
		createAutTabs();
		createTabs();
		tabbedPane.revalidate();
		ShipData.getInstance().setConventional(false);
		if(!startup){
			setTabbedPane(currenttab);
		}
	}
	
	/**
	 * Are we in conventional mode.
	 * @return Returns true when in conventional mode.
	 */
	public boolean getConventional(){
		return conventional;
	}
	
	/**
	 * Set filename in title.
	 * @param file File.
	 */
	public void setFileName(File file){
		if(file != null){
			setTitle(Settings.getInstance().
					getResourceBundle().getString("programTitle") 
					+ " - " + file.getName());
			mainMenu.getRecentMenu().insertItem(file.getAbsolutePath());
		} else {
			setTitle(Settings.getInstance().
					getResourceBundle().getString("programTitle") 
					+ " - " + Settings.getInstance().
					getResourceBundle().getString("untitled"));
		}
		
	}

	/**
	 * Update the main menu.
	 */
	public void updateMenu(){
		mainMenu.update();
	}
	
	/**
	 * Method to check if file changed and open file.
	 * @param openfile File object to open.
	 */
	public void openFile(File openfile){
		mainMenu.openFile(openfile);
	}
}
