//: /be/optiloading/gui/MainMenu.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2008 Tim Pintens. All rights reserved.
 */

package be.optiloading.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.sql.Time;
import java.text.NumberFormat;
import java.util.ResourceBundle;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.border.BevelBorder;
import javax.swing.border.EmptyBorder;

import be.optiloading.optimization.Lp;
import be.optiloading.optimization.Sa;
import be.optiloading.settings.Settings;

/**
 * ProgressPanel that is shown while algorithm is running.
 * 
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank"
 *         href="http://www.optiloading.be">www.optiloading.be</a>
 */

@SuppressWarnings("serial")
public class ProgressPanel extends JPanel{
	private ResourceBundle languageResource;
	private static ProgressPanel uniqueInstance;
	private JTextArea log;
	private JLabel timelabel;
	private long begintime;
	private TimeCounter counter;
	private NumberFormat formatter;
	private Lp lpproblem;
	private Sa saproblem;
	private JButton cancelButton;
	private boolean lp;
	private JProgressBar progressbar;
	
	/**
	 * Returns the uniqueInstance of the ProgressPanel object.
	 * 
	 * @return ProgressPanel object.
	 */
	public static ProgressPanel getInstance() {
		if (uniqueInstance == null) {
			uniqueInstance = new ProgressPanel();
		}
		return uniqueInstance;
	}
	
	// Private constructor.
	private ProgressPanel(){
		super(new BorderLayout());
		languageResource = Settings.getInstance().getResourceBundle();
		progressbar = new JProgressBar();
		progressbar.setPreferredSize(new Dimension(700,25));
		
		
		JPanel progressContainer = new JPanel();
		progressContainer.setLayout(new BoxLayout(progressContainer,BoxLayout.Y_AXIS));
		JPanel busyPanel = new JPanel(new BorderLayout());
		JLabel busy = new JLabel(languageResource.getString("busyMessage"));
		busyPanel.setBorder(new EmptyBorder(0,0,5,0));
		busyPanel.add(BorderLayout.CENTER, busy);
		JPanel logpanel = new JPanel(new BorderLayout());
		logpanel.setBorder(new EmptyBorder(5,0,0,0));
		logpanel.setPreferredSize(new Dimension(700,400));
		log = new JTextArea();
		log.setFont(new Font("monaco",Font.PLAIN,10));
		log.setEditable(false);
		JScrollPane scroll = new JScrollPane(log,
				JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
				JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		scroll.setBorder(new BevelBorder(BevelBorder.LOWERED));
		logpanel.add(BorderLayout.CENTER,scroll);
		JPanel timeandcancel = new JPanel(new BorderLayout());
		timeandcancel.setBorder(new EmptyBorder(5,0,0,0));
		timelabel = new JLabel();
		timelabel.setBorder(new EmptyBorder(5,0,5,0));
		cancelButton = new JButton(languageResource.getString("cancelLabel"));
		cancelButton.addActionListener(new ActionListener(){
			@SuppressWarnings("deprecation")
			public void actionPerformed(ActionEvent arg0) {
				if(lp){
					if(Settings.getInstance().getSolver() == Lp.GLPK){
						lpproblem.stop();
					} else {
						lpproblem.abort();
					}
				} else {
					saproblem.requestStop();
				}
				
				stopCounter();
				MainFrame.getInstance().unlock();
			}
		});
		timeandcancel.add(BorderLayout.WEST,timelabel);
		timeandcancel.add(BorderLayout.EAST,cancelButton);
		progressContainer.add(busyPanel);
		progressContainer.add(progressbar);
	
		add(BorderLayout.NORTH,progressContainer);
		add(BorderLayout.CENTER,logpanel);
		add(BorderLayout.SOUTH,timeandcancel);
		setBorder(new EmptyBorder(20,100,20,100));
		
	}
	
	/**
	 * Add msg to progresspanel.
	 * @param s String with message.
	 */
	public void addMsg(String s){
		Time time = new Time(System.currentTimeMillis());
		String timeString = "[" + time.toString() + "] ";
		log.append(timeString + "\t" + s +"\n");
		log.setCaretPosition(log.getDocument().getLength());
	}
	
	/**
	 * Start counter for LP problem.
	 * @param problem
	 */
	public void startCounter(Lp problem){
		this.lpproblem = problem;
		progressbar.setIndeterminate(true);
		lp = true;
		startCounter();
	}
	
	/**
	 * Start counter for SA problem.
	 * @param problem
	 */
	public void startCounter(Sa problem){
		this.saproblem = problem;
		progressbar.setIndeterminate(false);
		progressbar.setMinimum(0);
		progressbar.setMaximum(Settings.getInstance().getMaxIterations()
				*(Settings.getInstance().getNumberOfRestarts()+1));
		lp = false;
		startCounter();
	}
	
	/**
	 * Set progress to specified integer.
	 * @param val Integer value to set progress to.
	 */
	public void setProgress(int val){
		progressbar.setValue(val);
	}
	
	// Method to start counter
	private void startCounter(){
		begintime = System.currentTimeMillis();
		formatter = NumberFormat.getInstance();
        formatter.setMaximumFractionDigits(1);
        counter = new TimeCounter();
		counter.start();
	}
	
	/**
	 * Stop Counter
	 */
	public void stopCounter(){
		if(counter != null){
			counter.requestStop();
		}
	}
	
	/**
	 * Clear all text.
	 */
	public void clearProgress(){
		log.setText("");
	}
	
	// To count elapsed time
	private class TimeCounter extends Thread{
		private volatile boolean start = true;
		private Long elapsed;
		
		public TimeCounter(){
			elapsed = 0l;
		}
		
		public void run(){
			while(start){
				elapsed = System.currentTimeMillis() - begintime;
				float elapsedf = elapsed.floatValue()/1000;
				timelabel.setText(languageResource.getString("elapsedMessage")+
						" (s): "+formatter.format(elapsedf));
				uniqueInstance.repaint();
				try {
					Thread.sleep(100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
		
		// Stop counter
		public void requestStop(){
			start = false;
			if(lp){
				Log.getInstance().add(Log.LP,languageResource.getString("timeElapsed")
						+": "+String.valueOf(counter.getElapsed()));
			} else {
				Log.getInstance().add(Log.SA,languageResource.getString("timeElapsed")
						+": "+String.valueOf(counter.getElapsed()));
			}
			
		}
		 // Get elapsed time
		public float getElapsed(){
			return elapsed.floatValue()/1000;
		}
	}
	
}
