//: /be/optiloading/ship/ShipData.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2007 Tim Pintens. All rights reserved.
 */

package be.optiloading.ship;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.ResourceBundle;

import javax.swing.JOptionPane;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import be.optiloading.cargo.CargoList;
import be.optiloading.compatibility.CargoComp;
import be.optiloading.compatibility.SegregationComp;
import be.optiloading.compatibility.TankComp;
import be.optiloading.gui.Splash;
import be.optiloading.helpers.files.FileHelper;
import be.optiloading.settings.Settings;
import be.optiloading.tank.TankList;

/**
 * ShipData class which contains all ship information.
 * 
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank"
 *         href="http://www.optiloading.be">www.optiloading.be</a>
 */

public class ShipData {
	private CargoList cargoList;
	private CargoComp cargoCompList;
	private TankComp tankCompList;
	private SegregationComp segregationCompList;
	private TankList miscTankList;
	private TankList cargoTankList;
	private TankList ballastTankList;
	private StabilityData stabilityData;
	private StrengthData strengthData;
	private ArrayList<Float> sf;
	private ArrayList<Float> bm;

	private float begindist;
	private float enddist;
	private float lightship;
	private float lightshiplcg;
	private float lightshipvcg;
	private float lightshiptcg;
	private float floodtcg;
	private float floodvcg;
	private float floodlcg;
	private float loa;
	private float lbpp;
	private float maxdeadweight;
	private float h;
	private float lastdisplacement;
	private float lastdens;
	private float lastlcg;
	private static int intervals;
	private boolean conventional = false;
	
	private static ShipData uniqueInstance;
	private ResourceBundle languageResource;
	
	private Splash splash;
	
	public void setSplash(Splash splash) {
		this.splash = splash;
	}

	/**
	 * Get singleton instance.
	 * 
	 * @return ShipData object.
	 */
	public static ShipData getInstance(){
		if (uniqueInstance == null) {
			uniqueInstance = new ShipData();
		}
		return uniqueInstance;
	}
	
	//Singleton constructor
	private ShipData(){
		languageResource = Settings.getInstance().getResourceBundle();
	}
	
	/**
	 * Initialize general ship data.
	 * 
	 * @param xmlFile XML file containing the data.
	 */
	public void initializeShipData(String xmlFile){
		try {
			InputStream inputSource = this.getClass().
				getResource(xmlFile).openStream();
			URL xsdURL = this.getClass().getResource("/data/ship.xsd");
			if(FileHelper.checkFile(null, inputSource,xsdURL)){
				SAXParserFactory factory = SAXParserFactory.newInstance();
				factory.setValidating(true);
				factory.setNamespaceAware(false);
				
				SAXParser parser = factory.newSAXParser();
				inputSource = this.getClass().getResource(xmlFile).openStream();
				parser.parse(inputSource, new ShipDataHandler(this));
				inputSource.close();
			} else {
				invalidClose();
			}
		} catch (Exception e) {
			parseClose();
		}
		intervals = new Float(10*getLoa()).intValue();
	}
	
	//Handler for parsing the general ship data
	private class ShipDataHandler extends DefaultHandler{
		float current;
		boolean floatvalue;
		
		ShipData data;
		private StringBuffer buf;
		
		public ShipDataHandler(ShipData data){
			this.data = data;
			floatvalue = false;
		}

		public void startElement(String uri, String name, String qname,
				Attributes att) throws SAXException {
			if((qname.equals("lightship")) || (qname.equals("lightshiplcg")) || 
					(qname.equals("lightshipvcg")) || (qname.equals("lightshiptcg")) || 
					(qname.equals("loa")) || (qname.equals("lbpp")) || 
					(qname.equals("floodlcg")) || (qname.equals("floodtcg")) ||
					(qname.equals("begindist")) || (qname.equals("enddist")) ||
					(qname.equals("floodvcg")) || (qname.equals("deadweight"))){
				floatvalue = true;
			}
			buf = new StringBuffer();
		}
		
		public void characters(char[] ch, int start, int length){
			if(buf != null){
				buf.append(ch, start, length);
			}	
		}

		
		public void endElement(String uri, String name, String qname)
		throws SAXException {
			if(floatvalue == true){
				current = Float.valueOf(buf.toString());
				floatvalue = false;
			}
			if(qname.equals("lightship")){
				data.lightship = current;
			}
			if(qname.equals("lightshiplcg")){
				data.lightshiplcg = current;
			}
			if(qname.equals("lightshipvcg")){
				data.lightshipvcg = current;
			}
			if(qname.equals("lightshiptcg")){
				data.lightshiptcg  = current;
			}
			if(qname.equals("loa")){
				data.loa = current;
			}
			if(qname.equals("lbpp")){
				data.lbpp = current;
			}
			if(qname.equals("deadweight")){
				data.maxdeadweight = current;
			}
			if(qname.equals("floodtcg")){
				data.floodtcg = current;
			}
			if(qname.equals("floodvcg")){
				data.floodvcg = current;
			}
			if(qname.equals("floodlcg")){
				data.floodlcg = current;
			}
			if(qname.equals("begindist")){
				data.begindist = current;
			}
			if(qname.equals("enddist")){
				data.enddist = current;
			}
		}
	}
	
	/**
	 * Initialize stability data.
	 * 
	 * @param xmlHydrostaticFile XML file containing the hydrostatic data.
	 * @param xmlKnFile XML file containing the KN data.
	 */
	public void initializeStabilityData(String xmlHydrostaticFile,String xmlKnFile){
		try {
			URL xsdURL1 = this.getClass().getResource("/data/hydrostatic.xsd");
			URL xsdURL2 = this.getClass().getResource("/data/kn.xsd");
			if((FileHelper.checkFile(null, getClass().getResource(xmlHydrostaticFile)
					.openStream(),xsdURL1)) && (FileHelper.checkFile(null, getClass()
							.getResource(xmlKnFile).openStream(),xsdURL2))){
				stabilityData = new StabilityData(xmlHydrostaticFile, xmlKnFile);
			} else {
				invalidClose();
			}
		} catch (IOException e) {
			parseClose();
		}
	}
	
	/**
	 * Initialize strength data.
	 * 
	 * @param xmlLightshipFile File which contains lightship weight distribution.
	 * @param xmlBuoyancyFile File which contains buoyancy division.
	 */
	public void initializeStrengthData(String xmlLightshipFile,String xmlBuoyancyFile){
		lastdisplacement = 0f;
		lastlcg = 0f;
		lastdens = Settings.getInstance().getDensity();
		try {
			URL xsdURL1 = this.getClass().getResource("/data/lightship.xsd");
			URL xsdURL2 = this.getClass().getResource("/data/buoyancy.xsd");
			if((FileHelper.checkFile(null, getClass().getResource(xmlLightshipFile)
					.openStream(),xsdURL1)) && (FileHelper.checkFile(null, getClass()
							.getResource(xmlBuoyancyFile).openStream(),xsdURL2))){
				strengthData = new StrengthData(xmlLightshipFile, xmlBuoyancyFile);
			} else {
				invalidClose();
			}
		} catch (IOException e) {
			parseClose();
		}
		
	}
	
	/**
	 * Initialize misc tank data.
	 * 
	 * @param xmlFile XML file containing the data.
	 */
	public void initializeMiscTankList(String xmlFile){
		try {
			URL xsdURL = this.getClass().getResource("/data/tanks.xsd");
			if(FileHelper.checkFile(null, getClass().getResource(xmlFile)
					.openStream(),xsdURL)){
				miscTankList = new TankList(xmlFile, false);
			} else {
				invalidClose();
			}
		} catch (IOException e) {
			parseClose();
		}
		miscTankList.setDefaultMaxFactor(1f);
	}
	
	/**
	 * Get the TankList containing the misc tanks.
	 * 
	 * @return TankList with misc tanks.
	 */
	public TankList getMiscTankList(){
		return miscTankList;
	}
	
	/**
	 * Initialize cargo tank data.
	 * 
	 * @param xmlFile XML file containing the data.
	 */
	public void initializeCargoTankList(String xmlFile){
		segregationCompList = new SegregationComp();
		try {
			URL xsdURL = this.getClass().getResource("/data/tanks.xsd");
			if(FileHelper.checkFile(null, getClass().getResource(xmlFile)
					.openStream(),xsdURL)){
				cargoTankList = new TankList(xmlFile, true);
			} else {
				invalidClose();
			}
		} catch (IOException e) {
			parseClose();
		}
		cargoTankList.setDefaultMaxFactor(0.98f);
	}
	
	/**
	 * Get the TankList containing the cargo tanks.
	 * 
	 * @return TankList with cargo tanks.
	 */
	public TankList getCargoTankList(){
		return cargoTankList;
	}
	
	/**
	 * Initialize ballast tank data.
	 * 
	 * @param xmlFile XML file containing the data.
	 */
	public void initializeBallastTankList(String xmlFile){
		try {
			URL xsdURL = this.getClass().getResource("/data/tanks.xsd");
			if(FileHelper.checkFile(null, getClass().getResource(xmlFile)
					.openStream(),xsdURL)){
				ballastTankList = new TankList(xmlFile, false);
			}  else {
				invalidClose();
			}
		} catch (IOException e) {
			parseClose();
		}
		ballastTankList.setDefaultMaxFactor(1f);
	}
	
	/**
	 * Exit program when parse exception occurs. Ship data not properly loaded!
	 */ 
	public void parseClose(){
		String osName = System.getProperty("os.name");
		if(osName.equals("Linux") && (splash != null)){
			splash.toBack();
		}
		JOptionPane.showMessageDialog(null, 
				languageResource.getString("parsingExcption"),
				languageResource.getString("exceptionTitle"),
				JOptionPane.WARNING_MESSAGE);
		System.exit(1);
	}
	
	// Exit program when ship data file format is invalid
	private void invalidClose() {
		String osName = System.getProperty("os.name");
		if(osName.equals("Linux")){
			splash.toBack();
		}
		JOptionPane.showMessageDialog(null, Settings.getInstance()
				.getResourceBundle().getString("parsingExcption2"), Settings
				.getInstance().getResourceBundle().getString("exceptionTitle"),
				JOptionPane.WARNING_MESSAGE);
		System.exit(1);
	}
	
	/**
	 * Get the TankList containing the ballast tanks.
	 * 
	 * @return TankList with ballast tanks.
	 */
	public TankList getBallastList(){
		return ballastTankList;
	}
	
	/**
	 * Create CargoList.
	 */
	public void initializeCargoList(){
		cargoList = new CargoList();
	}
	
	/**
	 * getCargoList
	 * 
	 * @return CargoList
	 */
	public CargoList getCargoList(){
		return cargoList;
	}
	
	/**
	 * Create CargoCompList.
	 */
	public void initializeCargoCompList(){
		cargoCompList = new CargoComp();
	}
	
	/**
	 * getCargoCompList
	 * 
	 * @return CargoCompList
	 */
	public CargoComp getCargoCompList(){
		return cargoCompList;
	}
	
	/**
	 * getSegregationCompList
	 * 
	 * @return SegregationCompList
	 */
	public SegregationComp getSegregationCompList(){
		return segregationCompList;
	}
	
	/**
	 * Create TankCompList.
	 */
	public void initializeTankCompList(){
		tankCompList = new TankComp();
	}
	
	/**
	 * getTankCompList
	 * 
	 * @return CargoCompList
	 */
	public TankComp getTankCompList(){
		return tankCompList;
	}
	
	/**
	 * getLightship
	 * 
	 * @return Float with lighthip weight.
	 */
	public float getLightship() {
		return lightship;
	}
	
	/**
	 * getLightshiplcg
	 * 
	 * @return Float with lighthip lcg.
	 */
	public float getLightshiplcg() {
		return lightshiplcg;
	}
	
	/**
	 * getLightshipvcg
	 * 
	 * @return Float with lighthip vcg.
	 */
	public float getLightshipvcg() {
		return lightshipvcg;
	}
	
	/**
	 * getLightshiptcg
	 * 
	 * @return Float with lighthip tcg.
	 */
	public float getLightshiptcg() {
		return lightshiptcg;
	}
	
	/**
	 * getBegindist
	 * 
	 * @return Float with begin dist.
	 */
	public float  getBegindist() {
		return begindist;
	}
	
	/**
	 * getEnddist
	 * 
	 * @return Float with end dist.
	 */
	public float  getEnddist() {
		return enddist;
	}
	
	/**
	 * getLoa
	 * 
	 * @return Float with overall length.
	 */
	public float getLoa() {
		return loa;
	}
	
	/**
	 * getLbpp
	 * 
	 * @return Float with length between perpendiculars.
	 */
	public float getLbpp() {
		return lbpp;
	}
	
	/**
	 * getNumberofcargotanks.
	 * 
	 * @return Integer with number of cargo tanks.
	 */
	public int getNumberofcargotanks(){
		return cargoTankList.size();
	}
	
	/**
	 * getMaxcargovolume.
	 * 
	 * @return Float with max cargo volume.
	 */
	public float getMaxcargovolume(){
		return cargoTankList.getTotalMaxVolumes();
	}
	
	/**
	 * getDeadweightleft
	 * 
	 * @return Float with how much ca still be loaded.
	 */
	public float getDeadweightleft() {
		if(conventional){
			return (getMaxdeadweight() - getDeadweight());
		} else {
			return (getMaxdeadweight() - cargoList.getTotalWeight() 
					- miscTankList.getTotalWeight());
		}
		
	}
	
	/**
	 * getMaxdeadweight
	 * 
	 * @return Float with maximum deadweight.
	 */
	public float getMaxdeadweight(){
		return (((maxdeadweight+lightship) * (Settings.getInstance().getDensity()/1.025f)) 
				- lightship);
	}
	
	/**
	 * getMaxdeadweightinseawater
	 * 
	 * @return Float with maximum deadweight in sea water.
	 */
	public float getMaxdeadweightinseawater(){
		return maxdeadweight;
	}
	
	/**
	 * getDisplacement
	 * 
	 * @return Float with current displacement.
	 */
	public float getDisplacement(){
		return (lightship + miscTankList.getTotalWeight() 
				+ cargoTankList.getTotalWeight() +
				ballastTankList.getTotalWeight());
	}
	
	/**
	 * getDeadweight
	 * 
	 * @return Float with current deadweight.
	 */
	public float getDeadweight(){
		return (cargoTankList.getTotalWeight() +
				ballastTankList.getTotalWeight() +
				miscTankList.getTotalWeight());
	}
	
	/**
	 * getLcg()
	 * 
	 * @return Float with longitudinal center of gravity.
	 */
	public float getLcg(){
		float sum = 0;
		sum = cargoTankList.getLongMoment() + ballastTankList.getLongMoment() +
		miscTankList.getLongMoment() + (lightship*lightshiplcg);
		
		return (sum/getDisplacement());
	}
	
	/**
	 * getVcg()
	 * 
	 * @return Float with vertical center of gravity.
	 */
	public float getVcg(){
		float sum = 0;
		sum = cargoTankList.getVertMoment() + ballastTankList.getVertMoment() +
		miscTankList.getVertMoment() + (lightship*lightshipvcg);
		
		return (sum/getDisplacement());
	}
	
	/**
	 * getTcg()
	 * 
	 * @return Float with transversal center of gravity.
	 */
	public float getTcg(){
		float sum = 0;
		sum = cargoTankList.getTransMoment() + ballastTankList.getTransMoment() +
		miscTankList.getTransMoment()+ (lightship*lightshiptcg);
		
		return (sum/getDisplacement());
	}
	
	/**
	 * getFsc()
	 * 
	 * @return Float with free surface correction.
	 */
	public float getFsc(){
		float sum = 0;
		sum = cargoTankList.getFreeSurfaceMoment() + ballastTankList.getFreeSurfaceMoment() +
		miscTankList.getFreeSurfaceMoment();
		
		return (sum/getDisplacement());
	}
	
	/**
	 * getGm()
	 * 
	 * @return Float with initial metacenter position.
	 */
	public float getGm(){
		return (stabilityData.getKm()-getVcg()-getFsc());
	}
	
	/**
	 * getTrim()
	 * 
	 * @return Float with current trim.
	 */
	public float getTrim(){
		return ((getDisplacement()*(stabilityData.getLcb()-getLcg()))/
				(stabilityData.getMctc()*100));
	}
	
	/**
	 * getHeel()
	 * 
	 * @return Float with current heel.
	 */
	public float getHeel(){
		return (float) Math.toDegrees(Math.atan(getTcg()/getGm()));
	}
	
	/**
	 * getDaft()
	 * 
	 * @return Float with current draught aft.
	 */
	public float getDaft(){
		return (stabilityData.getTmd() + getTrim()*(stabilityData.getLcf()/lbpp));
	}
	
	/**
	 * getDfwd()
	 * 
	 * @return Float with current draught fwd.
	 */
	public float getDfwd(){
		return (stabilityData.getTmd() - getTrim()*((lbpp - stabilityData.getLcf())/lbpp));
	}
	
	/**
	 * getMindraught()
	 * 
	 * @return Float with minimum allowable draught.
	 */
	public float getMindraught(){
		return stabilityData.getMindraught();
	}
	
	/**
	 * getMaxdraught()
	 * 
	 * @return Float with maximum allowable draught.
	 */
	public float getMaxdraught(){
		return stabilityData.getMaxdraught();
	}

	/**
	 * getTmd()
	 * 
	 * @return Float with current true mean draught.
	 */
	public float getTmd() {
		return stabilityData.getTmd();
	}

	/**
	 * getTpc()
	 * 
	 * @return Float with current ton per centimeter.
	 */
	public float getTpc() {
		return stabilityData.getTpc();
	}

	/**
	 * getMctc()
	 * 
	 * @return Float with current moment to change trim one centimeter.
	 */
	public float getMctc() {
		return stabilityData.getMctc();
	}

	/**
	 * getLcb()
	 * 
	 * @return Float with current longitudinal center of buoyancy.
	 */
	public float getLcb() {
		return stabilityData.getLcb();
	}

	/**
	 * getLcf()
	 * 
	 * @return Float with current longitudinal center of floatation.
	 */
	public float getLcf() {
		return stabilityData.getLcf();
	}

	/**
	 * getKm()
	 * 
	 * @return Float with current keel-metacenter distance.
	 */
	public float getKm() {
		return stabilityData.getKm();
	}

	/**
	 * getKm()
	 * 
	 * @return 2D float array with current kn information.
	 */
	public float[][] getKn(){
		return stabilityData.getKn() ;
	}
	
	/**
	 * getDeadweight for a certain tmd.
	 * 
	 * @param tmd Float with desired tmd.
	 * @return Float with deadweight for desired tmd.
	 */
	public float getDeadweight(float tmd){
		float seadisp = stabilityData.getDisplacement(tmd);
		float currentdisp = seadisp * (1.025f/Settings.getInstance().getDensity());
		return (currentdisp - lightship - cargoList.getTotalWeight() - 
				miscTankList.getTotalWeight());
	}
	
	/**
	 * getCargoWeight()
	 * 
	 * @return Float with total cargo weight.
	 */
	public float getCargoWeight() {
		return cargoList.getTotalWeight();
	}
	
	/**
	 * getCargoVolume()
	 * 
	 * @return Float with total cargo volume.
	 */
	public float getCargoVolume(){
		return cargoList.getTotalVolume();
	}
	
	/**
	 * getFloodAngle()
	 * 
	 * @return Float with current angle of flooding.
	 */
	public float getFloodAngle(){
		float draughtatflood = interpolate(floodlcg,0,lbpp,getDaft(),getDfwd());
		return (float) Math.toDegrees(Math.atan((floodvcg-draughtatflood)/floodtcg));
	}
	
	/**
	 * getArea30()
	 * 
	 * @return Float with area under the GZ curve from 0°-30°.
	 */
	public float getArea30(){
		float sum = (float) - Math.abs(getTcg());
		for(int i = 1; i < 30;i++){
			int factor;
			if((i % 2) == 0){
				factor = 2;
			} else {
				factor = 4;
			}
			sum += factor * (getKn(i) - (getVcg() * Math.sin(Math.toRadians(i))) 
					- (getTcg() * Math.cos(Math.toRadians(i))));
		}
		sum += (float) (getKn(30) - (getVcg() * Math.sin(Math.toRadians(30))) 
				- (getTcg() * Math.cos(Math.toRadians(30))));
		return (float) (sum*(Math.toRadians(1)/3));
	}
	
	/**
	 * getArea40()
	 * 
	 * @return Float with area under the GZ curve from 0°-40°.
	 */
	public float getArea30to40(){
		float sum = (float) (getKn(30) - (getVcg() * Math.sin(Math.toRadians(30))) 
				- (getTcg() * Math.cos(Math.toRadians(30))));
		for(int i = 31; i < 40;i++){
			int factor;
			if((i % 2) == 0){
				factor = 2;
			} else {
				factor = 4;
			}
			sum += factor * (getKn(i) - (getVcg() * Math.sin(Math.toRadians(i))) 
					- (getTcg() * Math.cos(Math.toRadians(i))));
		}
		sum += (float) (getKn(40) - (getVcg() * Math.sin(Math.toRadians(40))) 
				- (getTcg() * Math.cos(Math.toRadians(40))));
		return (float) (sum*(Math.toRadians(1)/3));
	}
	
	/**
	 * getArea40()
	 * 
	 * @return Float with area under the GZ curve from 0°-40°.
	 */
	public float getArea40(){
		return (getArea30()+getArea30to40());
	}
	
	/**
	 * getMaxGZAngle()
	 * 
	 * @return Float with the angle where the top of the GZ curve is reached.
	 */
	public float getMaxGZAngle(){
		float maxgz = 0f;
		float maxgzangle = 0f;
		for(int i = 0; i < getKn().length; i++){
			float gz = (float) (getKn(getKn()[i][0]) 
					- (getVcg() * Math.sin(Math.toRadians(getKn()[i][0]))) 
					- (getTcg() * Math.cos(Math.toRadians(getKn()[i][0]))));
			if (gz>maxgz){
				maxgzangle = getKn()[i][0];
				maxgz = gz;
			}
		}
		return maxgzangle;
	}

	/**
	 * getAngleofLoll()
	 * 
	 * @return Float with angle of loll.
	 */
	public float getAngleofLoll(){
		float mingz = 0f;
		float mingzangle = 0f;
		int maxangle = new Float(stabilityData.getKn()[stabilityData.getKn().length-1][0]).intValue();
		for(float i = 0; i <= maxangle;i=i+0.01f){
			float gz = (float) (getKn(i) - (getVcg() * Math.sin(Math.toRadians(i))) 
			- (getTcg() * Math.cos(Math.toRadians(i))));
			if(gz < mingz){
				mingz = Math.abs(gz);
				mingzangle = i;
			}
			if(gz>0){
				break;
			}
		}
		return mingzangle;
	}
	
	/**
	 * getGZ30()
	 * 
	 * @return Float with GZ value at 30°.
	 */
	public float getGZ30(){
		return (float) (getKn(30) - (getVcg() * Math.sin(Math.toRadians(30))) 
				- (getTcg() * Math.cos(Math.toRadians(30))));
	}
	
	/**
	 * getKn for a certain angle.
	 * 
	 * @param angle for which you want the kn.
	 * @return Float with the kn value for the desired angle.
	 */
	private float getKn(float angle){
		float returnvalue = 0f;
		float[][] kn = stabilityData.getKn();
		for (int i = 0; i < kn.length; ++i) {
			if (angle < kn[i][0]) {
				returnvalue = interpolate(angle,kn[i-1][0],kn[i][0],kn[i-1][1],kn[i][1]);
				break;
			} else if (angle == kn[i][0]) {
				returnvalue = kn[i][1];
				break;
			}
		}
		return returnvalue;
	}
	
	/**
	 * getLoad()
	 * 
	 * @param dist Distance from APP.
	 * @return Floating point with load in ton/m.
	 */
	public float getLoad(float dist){
		return cargoTankList.getLoad(dist) + miscTankList.getLoad(dist) + 
		ballastTankList.getLoad(dist) + strengthData.getLightshipLoad(dist)
		+ strengthData.getBuoyancyLoad(dist, getLcb(), getLcg(), getLcf(), getDisplacement());
	}
	
	/**
	 * Get buoyancy force at specified distance.
	 * @param dist Distance in meters.
	 * @return Floating point value containing the buoyancy force in ton.
	 */
	public float getBuoyancy(float dist){
		return strengthData.getBuoyancyLoad(dist, getLcb(), getLcg(), getLcf(), getDisplacement());
	}
	
	/**
	 * Get downward force at specified distance.
	 * @param dist Distance in meters.
	 * @return Floating point value containing the downward force in ton.
	 */
	public float getWeight(float dist){
		return cargoTankList.getLoad(dist) + miscTankList.getLoad(dist) + 
		ballastTankList.getLoad(dist) + strengthData.getLightshipLoad(dist);
	}
	
	/**
	 * Get the corrected buoyancy for a specified section. The corrected buoyancy
	 * force is the buoyancy force plus the downward force for the light ship 
	 * and misc tanks.
	 * @param aft Aft of section in meters.
	 * @param fwd Fwd of section in meters.
	 * @param lcb Longitudinal center of buoyancy.
	 * @param lcg Longitudinal center of gravity.
	 * @param lcf Longitudinal center of floatation.
	 * @param disp Displacement.
	 * @return Correct buoyancy force of section in ton.
	 */
	public float getBuoymisc(float aft, float fwd, float lcb, float lcg, 
			float lcf, float disp){
		float length = fwd - aft;
		int numbersec = (int) (length*10);
		float seclength = length/numbersec;
		float total = 0f;
		for(int i = 0; i < numbersec;i++ ){
			total += (seclength*strengthData.getBuoyancyLoad(aft+(i*seclength), lcb, lcg, 
					lcf, disp));
			total +=  (seclength*(miscTankList.getLoad(aft+(i*seclength)))); 
			total +=  (seclength*(strengthData.getLightshipLoad(aft+(i*seclength))));
		}
		return total;
	}
	
	/**
	 * Get shear force at specified distance.
	 * @param dist Distance.
	 * @return Shear force in ton.
	 */
	public float getSF(float dist){
		checkSFandBM();
		float returnvalue = 0f;
		if(dist >getEnddist()){
			returnvalue = sf.get(sf.size()-1);
		}else{
			dist = dist - begindist; //From l form AP to real l
			int i = (int) (dist/h);
			float factor = (dist/h) - i;
			returnvalue = ((sf.get(i+1)-sf.get(i))*factor)+sf.get(i);
		}
		return returnvalue;
	}
	
	/**
	 * Get maximum shear force.
	 * @return Maximum shear force in ton.
	 */
	public float getMaxSF(){
		checkSFandBM();
		return Collections.max(sf);
	}
	
	/**
	 * Distance at which max shear force occurs.
	 * @return Distance in meters.
	 */
	public float getMaxSFLength(){
		float dist = (sf.indexOf(Collections.max(sf))*h)+begindist;
		return dist;
	}
	
	/**
	 * Get minimum shear force.
	 * @return Minimum shear force in ton.
	 */
	public float getMinSF(){
		checkSFandBM();
		return Collections.min(sf);
	}
	
	/**
	 * Distance at which min shear force occurs.
	 * @return Distance in meters.
	 */
	public float getMinSFLength(){
		float dist = (sf.indexOf(Collections.min(sf))*h)+begindist;
		return dist;
	}
	
	/**
	 * Get bending moment at specified distance.
	 * @param dist Distance.
	 * @return Bending moment in ton*meter.
	 */
	public float getBM(float dist){
		checkSFandBM();
		float returnvalue = 0f;
		if(dist >getEnddist()){
			returnvalue = bm.get(bm.size()-1);
		}else{
			dist = dist - begindist; //From l form AP to real l
			int i = (int) (dist/h);
			float factor = (dist/h) - i;
			returnvalue = ((bm.get(i+1)-bm.get(i))*factor)+bm.get(i);
		}
		return returnvalue;
	}
	
	/**
	 * Get maximum bending moment.
	 * @return Bending moment in ton*meter.
	 */
	public float getMaxBM(){
		checkSFandBM();
		return Collections.max(bm);
	}
	
	/**
	 * Distance at which max bending moment occurs.
	 * @return Distance in meters.
	 */
	public float getMaxBMLength(){
		float dist = (bm.indexOf(Collections.max(bm))*h)+begindist;
		return dist;
	}
	
	/**
	 * Get minimum bending moment.
	 * @return Bending moment in ton*meter.
	 */
	public float getMinBM(){
		checkSFandBM();
		return Collections.min(bm);
	}
	
	/**
	 * Distance at which min bending moment occurs.
	 * @return Distance in meters.
	 */
	public float getMinBMLength(){
		float dist = (bm.indexOf(Collections.min(bm))*h)+begindist;
		return dist;
	}
	
	/**
	 * Get shear force or bending moment at specified distance.
	 * @param particular SF or BM.
	 * @param dist Distance in meters.
	 * @return SF in ton or BM in ton*meter.
	 */
	public float getStrength(ShipParticular particular,float dist){
		float returnvalue = 0;
		switch(particular){
		case SF: returnvalue = this.getSF(dist); break;
		case BM: returnvalue = this.getBM(dist) ; break;
		}
		return returnvalue;
	}
	
	/**
	 * Get the absolute max SF
	 * @return Max shear force in ton.
	 */
	public float getAbsMaxSF(){
		float returnvalue = 0f;
		if(getMaxSF()>Math.abs(getMinSF())){
			returnvalue = getMaxSF();
		} else {
			returnvalue = getMinSF();
		}
		return returnvalue;
	}
	
	/**
	 * Get the absolute max BM
	 * @return Max bending moment in ton*meter.
	 */
	public float getAbsMaxBM(){
		float returnvalue = 0f;
		if(getMaxBM()>Math.abs(getMinBM())){
			returnvalue = getMaxBM();
		} else {
			returnvalue = getMinBM();
		}
		return returnvalue;
	}
	
	/**
	 * Get distance at which absolute max SF occurs.
	 * @return Distance in meters.
	 */
	public float getAbsMaxSFLength(){
		float returnvalue = 0f;
		if(getMaxSF()>Math.abs(getMinSF())){
			returnvalue = getMaxSFLength();
		} else {
			returnvalue = getMinSFLength();
		}
		return returnvalue;
	}
	
	/**
	 * Get distance at which absolute max BM occurs.
	 * @return Distance in meters.
	 */
	public float getAbsMaxBMLength(){
		float returnvalue = 0f;
		if(getMaxBM()>Math.abs(getMinBM())){
			returnvalue = getMaxBMLength();
		} else {
			returnvalue = getMinBMLength();
		}
		return returnvalue;
	}
	
	// Check if SF and BM values are up to date
	private void checkSFandBM(){
		if((getDisplacement() != lastdisplacement) || (getLcg() != lastlcg) ||
				(Settings.getInstance().getDensity() != lastdens)){
			calculateSFandBM();
		}
	}
	
	/**
	 * Calculate SF and BM values.
	 */
	public void calculateSFandBM(){
		lastdisplacement = getDisplacement();
		lastlcg = getLcg();
		lastdens = Settings.getInstance().getDensity();
		
		sf = new ArrayList<Float>();
		bm = new ArrayList<Float>();

		sf.add(0f);
		bm.add(0f);
		
		h = getLoa()/intervals;
		float h3 = h/12;
		float h32 = h/6;
		for(int i = 1; i < (intervals+1);i++){
			float dist = (i*h) + begindist;
			float sfdist1 = h3 * (getLoad(dist-h)+4*getLoad(dist-((3*h)/4)) +getLoad(dist-(h/2)));
			float sfdist2 = h3 * (getLoad(dist-(h/2)) + 4*getLoad(dist-(h/4))+getLoad(dist));
			float sfdist = (sfdist1+sfdist2) + sf.get(i-1);
			sf.add(sfdist);
			float bmdist = h32 * (sf.get(i-1)+4*sfdist1+ sf.get(i));
			bmdist = bmdist + bm.get(i-1);
			bm.add(bmdist);
		}
	}
	
	/**
	 * Get stabilitydata object.
	 * @return StabilityData object (singleton).
	 */
	public StabilityData getStabilityData(){
		return stabilityData;
	}
	
	/**
	 * Correct trim.
	 * @param ballast True if correction has to be made with ballast.
	 */
	public void correctTrim(boolean ballast){
		if(ballast){
			ballastTankList.emptyTanks(false);
			ballastTankList.correctTrim(Settings.getInstance().getDesiredTrim(),
					getMaxBallast());
		}
	}
	
	/**
	 * Correct heel.
	 * @param ballast True if correction has to be made with ballast.
	 */
	public void correctHeel(boolean ballast){
		if(ballast){
			ballastTankList.emptyTanks(false);
			ballastTankList.correctHeel(Settings.getInstance().getDesiredHeel()
					,getMaxBallast());
		}
	}
	
	/**
	 * Correct trim and heel.
	 * @param ballast True if correction has to be made with ballast.
	 */
	public void correctBoth(boolean ballast){
		if(ballast){
			ballastTankList.emptyTanks(false);
			ballastTankList.correctBoth(Settings.getInstance().getDesiredTrim(),
					Settings.getInstance().getDesiredHeel(),getMaxBallast());
		}
	}
	
	/**
	 * Set to conventional mode.
	 * @param value True if conventional.
	 */
	public void setConventional(boolean value){
		conventional = value;
	}
	
	/**
	 * Returns the maximum amount of ballast.
	 * @return Float with the maximum amount of ballast.
	 */
	public float getMaxBallast(){
		return (ShipData.getInstance().getMaxdeadweight()
				- ShipData.getInstance().getMiscTankList().getTotalWeight()
				- ShipData.getInstance().getCargoTankList().getTotalWeight());
		
	}
	
	/**
	 * Get ship data.
	 * 
	 * @param particular for which you want the data.
	 * @return Float with the value of the desired particular.
	 */
	public float getShipParticular(ShipParticular particular){
		float returnvalue = 0;
		
		//case statement on the enum ShipParticular
		switch(particular){
			case DISPLACEMENT: returnvalue = this.getDisplacement(); break;
			case LIGHTSHIP: returnvalue = lightship ; break;
			case LIGHTSHIP_LCG: returnvalue = lightshiplcg ; break;
			case LIGHTSHIP_VCG: returnvalue = lightshipvcg ; break;
			case LIGHTSHIP_TCG: returnvalue = lightshiptcg ; break;
			case MAX_DEADWEIGHT: returnvalue = this.getMaxdeadweight() ; break;
			case MAX_CARGO_VOLUME: returnvalue = this.getMaxcargovolume() ; break;
			case HEEL: returnvalue = this.getHeel() ; break;
			case MAX_DRAUGHT: returnvalue = stabilityData.getMaxdraught() ; break;
			case MIN_DRAUGHT: returnvalue = stabilityData.getMindraught() ; break;
			case LOA: returnvalue = loa ; break;
			case LBPP: returnvalue = lbpp ; break;
			case NUMBEROFCARGOTANKS: returnvalue = cargoTankList.size(); break;
			case NUMBEROFBALLASTTANKS: returnvalue = ballastTankList.size(); break;
			case NUMBEROFMISCTANKS: returnvalue = miscTankList.size(); break;
			case TMD: returnvalue = stabilityData.getTmd(); break;
			case LCF: returnvalue = stabilityData.getLcf(); break;
			case LCB: returnvalue = stabilityData.getLcb(); break;
			case LCG: returnvalue = this.getLcg(); break;
			case MCTC: returnvalue = stabilityData.getMctc(); break;
			case TPC: returnvalue = stabilityData.getTpc(); break;
			case FSC: returnvalue = this.getFsc(); break;
			case KM: returnvalue = stabilityData.getKm(); break;
			case VCG: returnvalue = this.getVcg(); break;
			case TCG: returnvalue = this.getTcg(); break;
			case GM: returnvalue = this.getGm(); break;
			case TRIM: returnvalue = this.getTrim(); break;
			case DFWD: returnvalue = this.getDfwd(); break;
			case DAFT: returnvalue = this.getDaft(); break;
			case DENSITY: returnvalue = Settings.getInstance().getDensity(); break;
			case DEADWEIGHT: returnvalue = this.getDeadweight(); break;
			case DEADWEIGHTLEFT: returnvalue = this.getDeadweightleft(); break;
			case CARGO_WEIGHT: 
				if(conventional){
					returnvalue = cargoTankList.getTotalWeight();
				} else {
					returnvalue = cargoList.getTotalWeight();
				}
				break;
			case CARGO_VOLUME: 
				if(conventional){
					returnvalue = cargoTankList.getTotalVolumes();
				} else {
					returnvalue = cargoList.getTotalVolume();
				}
				break;
			case ANGLE_OF_FLOODING: returnvalue = this.getFloodAngle(); break;
			case ANGLE_OF_LOLL: returnvalue = this.getAngleofLoll(); break;
			case AREA_30: returnvalue = this.getArea30(); break;
			case AREA_40: returnvalue = this.getArea40(); break;
			case AREA_30TO40: returnvalue = this.getArea30to40(); break;
			case MAX_GZ_ANGLE: returnvalue = this.getMaxGZAngle(); break;
			case GZ_30: returnvalue = this.getGZ30(); break;
			case BEGIN_DIST: returnvalue = this.getBegindist(); break;
			case END_DIST: returnvalue = this.getEnddist(); break;
			case MAX_BM: returnvalue = this.getMaxBM(); break;
			case MIN_BM: returnvalue = this.getMinBM(); break;
			case MAX_SF: returnvalue = this.getMaxSF(); break;
			case MIN_SF: returnvalue = this.getMinSF(); break;
			case MAX_BM_DIST: returnvalue = this.getMaxBMLength(); break;
			case MIN_BM_DIST: returnvalue = this.getMinBMLength(); break;
			case MAX_SF_DIST: returnvalue = this.getMaxSFLength(); break;
			case MIN_SF_DIST: returnvalue = this.getMinSFLength(); break;
			case ABS_MAX_SF: returnvalue = this.getAbsMaxSF(); break;
			case ABS_MAX_BM: returnvalue = this.getAbsMaxBM(); break;
			case ABS_MAX_SF_DIST: returnvalue = this.getAbsMaxSFLength(); break;
			case ABS_MAX_BM_DIST: returnvalue = this.getAbsMaxBMLength(); break;
			case BALLAST_VOLUME: returnvalue = ballastTankList.getTotalVolumes(); break;
			case BALLAST_WEIGHT: returnvalue = ballastTankList.getTotalWeight(); break;
			case CARGO_VOLUME_LEFT: 
				if(conventional){
					returnvalue = cargoTankList.getTotalMaxVolumes() - cargoTankList.getTotalVolumes();
				} else {
					returnvalue = cargoTankList.getTotalMaxVolumes() - cargoList.getTotalVolume();
				}
				break;
		}
		return returnvalue;
	}
	
	// Method used to interpolate between value
	// refnum is the number for which we want the value
	// refnum is between refsmall and reflarge which correspond to solsmall and
	// solllargo
	private float interpolate(float refnum, float refsmall,
			float reflarge, float solsmall, float sollarge) {
		float x = refnum - refsmall;
		float y = reflarge - refsmall;
		float z = sollarge - solsmall;
		float solution = (x / y) * z;
		return (solution + solsmall);
	}
}
