//: /be/optiloading/ship/StabilityData.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2007 Tim Pintens. All rights reserved.
 */

package be.optiloading.ship;

import java.io.InputStream;
import java.util.ArrayList;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import be.optiloading.settings.Settings;

/**
 * StabilityData class which contains all stability information. 
 * Holds the cross curves and hydrostatic table.
 * 
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank"
 *         href="http://www.optiloading.be">www.optiloading.be</a>
 */

public class StabilityData {
	private float tmd;
	private float tpc;
	private float mctc;
	private float lcb;
	private float lcf;
	private float km;
	private boolean dontcheck;
	
	private float[][] kn;
	
	private ArrayList<HydrostaticRecord> hydrostaticData;
	private ArrayList<KnRecord> knData;
	
	private float currentdisp;
	private float currentdens;
	
	/**
	 * Constructor of the HydrostaticData. Builds up the ArrayList<HydrostaticRecord>.
	 * 
	 * @param xmlHydrostaticFile File which contains hydrostatic information.
	 * @param xmlKnFile File which contains cross curves information.
	 */
	public StabilityData(String xmlHydrostaticFile,String xmlKnFile){
		dontcheck = false;
		initializeHydrostaticData(xmlHydrostaticFile);
		initializeKnData(xmlKnFile);
		currentdisp = 0f;
		currentdens = Settings.getInstance().getDensity();
	}
	
	//Initialize hydrostatic data
	private void initializeHydrostaticData(String xmlFile){
		hydrostaticData = new ArrayList<HydrostaticRecord>();
		try {
			InputStream inputSource = this.getClass().
			getResource(xmlFile).openStream();
			
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.setValidating(true);
			factory.setNamespaceAware(false);
			
			SAXParser parser = factory.newSAXParser();
			parser.parse(inputSource, new HydrostaticHandler(hydrostaticData));
		
		} catch (Exception e) {
			e.printStackTrace();
			ShipData.getInstance().parseClose();
		}
	}
	
	//Initialize Kn data
	private void initializeKnData(String xmlFile){
		knData = new ArrayList<KnRecord>();
		try {
			InputStream inputSource = this.getClass().
			getResource(xmlFile).openStream();
			
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.setValidating(true);
			factory.setNamespaceAware(false);
			
			SAXParser parser = factory.newSAXParser();
			parser.parse(inputSource, new KnHandler(knData));
		
		} catch (Exception e) {
			e.printStackTrace();
			ShipData.getInstance().parseClose();
		}
	}
	
	/**
	 * Set the displacement.
	 * 
	 * @param displacement Change current displacement.
	 */
	public void setDisplacement(float displacement){
		//Convert displacement to displacement in sea water.
		currentdisp =  ShipData.getInstance().getDisplacement();
		currentdens = Settings.getInstance().getDensity();
		float seadisplacement = displacement * (1.025f/currentdens);
		calculateHydrostaticData(seadisplacement);
		calculateKnData(seadisplacement);
	}
	
	//Calculate the other values for a specific displacement.
	private void calculateHydrostaticData(float displacement){
		for (int i = 0; i < hydrostaticData.size(); ++i) {
			HydrostaticRecord current = hydrostaticData.get(i);

			// If the displacement of the object is greater then we've gone past the
			// desired volume
			if (displacement < current.getDisplacement()) {
				// Get the previous object to interpolate
				HydrostaticRecord previous = hydrostaticData.get(i - 1);
				tmd = interpolate(displacement, previous.getDisplacement(), 
						current.getDisplacement(), previous.getTmd(), current.getTmd());
				tpc = interpolate(displacement, previous.getDisplacement(), 
						current.getDisplacement(), previous.getTpc(), current.getTpc());
				mctc = interpolate(displacement, previous.getDisplacement(), 
						current.getDisplacement(), previous.getMctc(), current.getMctc());
				lcb = interpolate(displacement, previous.getDisplacement(), 
						current.getDisplacement(), previous.getLcb(), current.getLcb());
				lcf = interpolate(displacement, previous.getDisplacement(), 
						current.getDisplacement(), previous.getLcf(), current.getLcf());
				km = interpolate(displacement, previous.getDisplacement(), 
						current.getDisplacement(), previous.getKm(), current.getKm());
				break;

				// The desired displacement is equal to the objects volume so we found
				// the object
			} else if (displacement == current.getDisplacement()) {
				tmd = current.getTmd();
				tpc = current.getTpc();
				mctc = current.getMctc();
				lcb = current.getLcb();
				lcf = current.getLcf();
				km = current.getKm();
				break;
			}
		}
	}
	
	//Calculate the other values for a specific displacement.	
	private void calculateKnData(float displacement){
		for (int i = 0; i < knData.size(); ++i) {
			KnRecord current = knData.get(i);

			// If the displacement of the object is greater then we've gone past the
			// desired volume
			if (displacement < current.getDisplacement()) {
				// Get the previous object to interpolate
				KnRecord previous = knData.get(i - 1);
				kn = new float[current.getData().length][2];
				
				for(int j = 0; j < current.getData().length;j++){
					kn[j][0] = current.getData()[j][0];
					kn[j][1] = interpolate(displacement, previous.getDisplacement(), 
							current.getDisplacement(), previous.getData()[j][1], 
							current.getData()[j][1]);
				}
				
				break;

				// The desired displacement is equal to the objects volume so we found
				// the object
			} else if (displacement == current.getDisplacement()) {
				kn = current.getData();
				break;
			}
		}
	}
	
	//SAX document handler to handle the callback.
	private class HydrostaticHandler extends DefaultHandler{
		private float tmd;
		private float displacement;
		private float tpc;
		private float mctc;
		private float lcb;
		private float lcf;
		private float km;
		
		float current;
		boolean floatvalue;
		private StringBuffer buf;
		
		ArrayList<HydrostaticRecord> data;
		
		public HydrostaticHandler(ArrayList<HydrostaticRecord> data){
			this.data = data;
			floatvalue = false;
		}

		public void startElement(String uri, String name, String qname,
				Attributes att) throws SAXException {
			if((qname.equals("tmd")) || (qname.equals("displacement")) || 
					(qname.equals("tpc")) || (qname.equals("mctc")) || 
					(qname.equals("lcb")) || (qname.equals("lcf"))
					|| (qname.equals("km"))){
				floatvalue = true;
			}
			buf = new StringBuffer();
		}
		
		public void characters(char[] ch, int start, int length){
			if(buf != null){
				buf.append(ch, start, length);
			}	
		}

		public void endElement(String uri, String name, String qname)
		throws SAXException {
			if(floatvalue == true){
				current = Float.valueOf(buf.toString());
				floatvalue = false;
			}
			if(qname.equals("record")){
				data.add(new HydrostaticRecord(tmd,displacement,tpc,mctc,
						lcb,lcf,km));
			}
			if(qname.equals("tmd")){
				tmd = current;
			}
			if(qname.equals("displacement")){
				displacement = current;
			}
			if(qname.equals("tpc")){
				tpc = current;
			}
			if(qname.equals("mctc")){
				mctc = current;
			}
			if(qname.equals("lcb")){
				lcb = current;
			}
			if(qname.equals("lcf")){
				lcf = current;
			}
			if(qname.equals("km")){
				km = current;
			}
		}
	}
	
	//SAX document handler to handle the callback.
	private class KnHandler extends DefaultHandler{
		private float displacement;
		private float[][] recorddata;
		private String[] valueNames;
		private float[] valueNumbers;
		
		float current;
		boolean floatvalue;
		
		ArrayList<KnRecord> data;
		
		private StringBuffer buf;
		
		public KnHandler(ArrayList<KnRecord> data){
			this.data = data;
			floatvalue = false;
		}

		public void startElement(String uri, String name, String qname,
				Attributes att) throws SAXException {
			if(qname.equals("crosscurves")){
				return;
			}
			
			if(qname.equals("kn")){
				valueNames = new String[att.getLength()];
				valueNumbers = new float[att.getLength()];
				for(int i = 0; i< att.getLength(); i++){
					valueNames[i] = att.getQName(i);
					valueNumbers[i] = Float.valueOf(att.getValue(i));
				}
				return;
			}
			
			if(qname.equals("record")){
				recorddata = new float[valueNumbers.length][2];
				for(int i = 0;  i < valueNumbers.length;i++ ){
					recorddata[i][0] = valueNumbers[i];
				}
				return;
			}
			
			if(qname.equals("displacement")){
				floatvalue = true;
				buf = new StringBuffer();
				return;
			}
			
			for(int i = 0; i < valueNames.length; i++){
				if(qname.equals(valueNames[i])){
					floatvalue = true;
					buf = new StringBuffer();
					return;
				}
			}
		}
		
		public void characters(char[] ch, int start, int length){
			if(buf != null){
				buf.append(ch, start, length);
			}	
		}

		public void endElement(String uri, String name, String qname)
		throws SAXException {
			if(floatvalue == true){
				current = Float.valueOf(buf.toString());
				floatvalue = false;
			}
			if(qname.equals("record")){
				data.add(new KnRecord(displacement, recorddata));
				return;
			}
			
			if(qname.equals("displacement")){
				displacement = current;
				return;
			}
			
			for(int i = 0; i < valueNames.length; i++){
				if(qname.equals(valueNames[i])){
					recorddata[i][1] = current;
					return;
				}
			}
			
		}
	}
	
	//Class with holds a row of the hydrostatic table.
	private class HydrostaticRecord{
		private float tmd;
		private float displacement;
		private float tpc;
		private float mctc;
		private float lcb;
		private float lcf;
		private float km;
		
		public HydrostaticRecord(float tmd, float displacement, float tpc,
				float mctc,float lcb,float lcf,float km){
			this.tmd = tmd;
			this.displacement = displacement;
			this.tpc = tpc;
			this.mctc = mctc;
			this.lcb = lcb;
			this.lcf = lcf;
			this.km = km;
		}

		public float getTmd() {
			return tmd;
		}

		public float getDisplacement() {
			return displacement;
		}

		public float getTpc() {
			return tpc;
		}

		public float getMctc() {
			return mctc;
		}

		public float getLcb() {
			return lcb;
		}

		public float getLcf() {
			return lcf;
		}

		public float getKm() {
			return km;
		}
	}
	
	//Class with holds a row of the cross curves table.
	private class KnRecord{
		float displacement;
		float[][] data;
		
		public KnRecord(float displacement, float[][] data){
			this.displacement = displacement;
			this.data = data;
		}

		public float getDisplacement() {
			return displacement;
		}

		public float[][] getData() {
			return data;
		}
	}
	
	// Method used to interpolate between value
	// refnum is the number for which we want the value
	// refnum is between refsmall and reflarge which correspond to solsmall and
	// solllargo
	private float interpolate(float refnum, float refsmall,
			float reflarge, float solsmall, float sollarge) {
		float x = refnum - refsmall;
		float y = reflarge - refsmall;
		float z = sollarge - solsmall;
		float solution = (x / y) * z;
		return (solution + solsmall);
	}

	/**
	 * getTmd
	 * 
	 * @return Float with current true mean draught.
	 */
	public float getTmd() {
		checkDisplacement();
		return tmd;
	}
	
	/**
	 * getDisplacement for a specific tmd.
	 * 
	 * @param truemd Float with the tmd.
	 * @return Float with displacement for the tmd.
	 */
	public float getDisplacement(float truemd) {
		float disp = 0f;
		for (int i = 0; i < hydrostaticData.size(); ++i) {
			HydrostaticRecord current = hydrostaticData.get(i);

			// If the displacement of the object is greater then we've gone past the
			// desired volume
			if (truemd < current.getTmd()) {
				// Get the previous object to interpolate
				HydrostaticRecord previous = hydrostaticData.get(i - 1);
				disp = interpolate(truemd, previous.getTmd(), current.getTmd(), 
						previous.getDisplacement(), current.getDisplacement());
				break;
				// The desired displacement is equal to the objects volume so we found
				// the object
			} else if (truemd == current.getTmd()) {
				disp = current.getDisplacement();
				break;
			}
		}
		return disp;
	}

	/**
	 * getTpc
	 * 
	 * @return Float with current ton per centimeter immersion.
	 */
	public float getTpc() {
		checkDisplacement();
		return tpc;
	}

	/**
	 * getMctc
	 * 
	 * @return Float with current moment to change trim 1 cm.
	 */
	public float getMctc() {
		checkDisplacement();
		return mctc;
	}

	/**
	 * getLcb
	 * 
	 * @return Float with longitudinal center of buoyancy.
	 */
	public float getLcb() {
		checkDisplacement();
		return lcb;
	}

	/**
	 * getLcf
	 * 
	 * @return Float with longitudinal center of floatation.
	 */
	public float getLcf() {
		checkDisplacement();
		return lcf;
	}

	/**
	 * getKm
	 * 
	 * @return Float with keel-metacenter height.
	 */
	public float getKm() {
		checkDisplacement();
		return km;
	}
	
	/**
	 * getMindraught
	 * 
	 * @return Float with minimum true mean draught.
	 */
	public float getMindraught(){
		checkDisplacement();
		return hydrostaticData.get(0).getTmd();
	}
	
	/**
	 * getMaxdraught
	 * 
	 * @return Float with maximum true mean draught.
	 */
	public float getMaxdraught(){
		checkDisplacement();
		return hydrostaticData.get(hydrostaticData.size()-1).getTmd();
	}
	
	/**
	 * getKn()
	 * 
	 * @return 2D array containing kn data.
	 */
	public float[][] getKn(){
		checkDisplacement();
		return kn;
	}
	
	private void checkDisplacement(){
		if(((currentdisp != ShipData.getInstance().getDisplacement())||
				(currentdens != Settings.getInstance().getDensity())) &&
				(dontcheck == false)){
			setDisplacement(ShipData.getInstance().getDisplacement());
		}
	}

	public void setDontcheck(boolean dontcheck) {
		this.dontcheck = dontcheck;
	}
}
