//: /be/optiloading/ship/StrengthData.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2007 Tim Pintens. All rights reserved.
 */

package be.optiloading.ship;

import java.io.InputStream;
import java.text.NumberFormat;
import java.util.ArrayList;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * StrengthData class which contains all strength information. 
 * Used for drawing the shear force and bending moments curve.
 * 
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank"
 *         href="http://www.optiloading.be">www.optiloading.be</a>
 */

public class StrengthData {
	private ArrayList<Section> lightshipList;
	private ArrayList<Section> buoyancyListStat;
	private ArrayList<Section> buoyancyList;
	private float displacement;
	private float lcb;
	private float lcg;
	private float lcf;
	
	/**
	 * Test procedure
	 * @param args Arguments from command line.
	 */
	public static void main(String[] args){
		StrengthData data = new StrengthData("/data/lightship.xml", "/data/buoyancy.xml");
		NumberFormat formatter = NumberFormat.getInstance();
		formatter.setMaximumFractionDigits(2);
		System.out.print(formatter.format(-data.getBuoyancyLoad(25.5f, 42.794f, 42.794f, 39.95f ,2525.754f)));																	
		System.exit(0);
	}
	
	/**
	 * Construtor.
	 * 
	 * @param xmlLightshipFile File which contains lightship weight distribution.
	 * @param xmlBuoyancyFile File which contains buoyancy division.
	 */
	public StrengthData(String xmlLightshipFile, String xmlBuoyancyFile){
		initializeLightship(xmlLightshipFile);
		initializeBuoyancy(xmlBuoyancyFile);
	}
	
	//Initialize lightship data
	private void initializeLightship(String xmlFile){
		lightshipList = new ArrayList<Section>();
		try {
			InputStream inputSource = this.getClass().
			getResource(xmlFile).openStream();
			
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.setValidating(true);
			factory.setNamespaceAware(false);
			
			SAXParser parser = factory.newSAXParser();
			parser.parse(inputSource, new LightShipHandler(lightshipList));
		
		} catch (Exception e) {
			ShipData.getInstance().parseClose();
		}
	}
	
	//Initialize lightship data
	private void initializeBuoyancy(String xmlFile){
		buoyancyListStat = new ArrayList<Section>();
		try {
			InputStream inputSource = this.getClass().
			getResource(xmlFile).openStream();
			
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.setValidating(true);
			factory.setNamespaceAware(false);
			
			SAXParser parser = factory.newSAXParser();
			parser.parse(inputSource, new BuoyancyHandler(buoyancyListStat));
		
		} catch (Exception e) {
			ShipData.getInstance().parseClose();
		}
	}
	
	/**
	 * Get the load at a certain dist from the APP.
	 * 
	 * @param dist Distance from the APP.
	 * @return Floating point with the load in ton/m.
	 */
	public float getLightshipLoad(float dist){
		float returnvalue = 0f;
		
		for(Section x: lightshipList){
			if ((x.getAftBound() <= dist) && (dist <= x.getFwdBound())){
				returnvalue = interpolate(dist,x.getAftBound(),x.getFwdBound(),
						x.getAftLoad(),x.getFwdLoad());
				break;
			}
		}
		
		return returnvalue;
	}
	
	/**
	 * Get the buoyancy load at a certain dist from the APP.
	 * 
	 * @param dist Distance from the APP.
	 * @param lcb Current Longitudinal Center of Buoyancy.
	 * @param lcg Current Longitudinal Center of Gravity.
	 * @param displacement Current displacement.
	 * @return Floating point with the load in ton/m.
	 */
	public float getBuoyancyLoad(float dist, float lcb, float lcg, float lcf,float displacement){
		if((this.displacement != displacement) || (this.lcb != lcb)
				|| (this.lcg != lcg) || (this.lcf != lcf)){
			this.displacement = displacement;
			this.lcb = lcb;
			this.lcg = lcg;
			this.lcf = lcf;
			calculateBuoyancyLoad();
		}
		float returnvalue = 0f;
		for(Section x: buoyancyList){
			if ((x.getAftBound() <= dist) && (dist <= x.getFwdBound())){
				returnvalue = interpolate(dist,x.getAftBound(),x.getFwdBound(),
						x.getAftLoad(),x.getFwdLoad());
				break;
			}
		}
		return - returnvalue;
	}
	
	//Private method to calculate initial buoyancy dist at aft-mid and fwd-mid section
	private void calculateBuoyancyLoad(){
		buoyancyList = new ArrayList<Section>();
		int smallSectionIndex = 0;
		int lcfSection = 0;
		float dist1 = 0;
		float dist2 = 0;
		boolean fwd = true;
		for(Section x: buoyancyListStat){
			if(x.getFactor() == 1f){
				float distAft = lcf - x.getAftBound();
				float distFwd = x.getFwdBound() - lcf;
				if(distAft < distFwd){
					float fbound = (x.getAftBound() + (2*distAft));
					dist1 = distAft;
					dist2 = x.getFwdBound() - fbound;
					lcfSection = buoyancyList.size();
					buoyancyList.add( new Section(x.getAftBound(),fbound,1f));
					smallSectionIndex = buoyancyList.size();
					fwd = true;
					buoyancyList.add(new Section(fbound,x.getFwdBound(),1f));
				} else {
					float abound = (x.getFwdBound()-(2*distFwd));
					dist1 = distFwd;
					dist2 = abound - x.getAftBound();
					smallSectionIndex = buoyancyList.size();
					fwd = false;
					buoyancyList.add(new Section(x.getAftBound(),abound,1f));
					lcfSection = buoyancyList.size();
					buoyancyList.add(new Section(abound,x.getFwdBound(),1f));
				}
			} else {
				buoyancyList.add(x);
			}
		}
		
		// Trim = 0
		float totalmoment = 0f;
		for (Section x: buoyancyList){
			totalmoment += x.getFactor() * x.getLength() * x.getCG();
		}
		float p = (lcb*displacement)/totalmoment;

		
		// Different Trim
		ArrayList<Integer> factorList = new ArrayList<Integer>(buoyancyList.size());
		ArrayList<Float> factorList2 = new ArrayList<Float>(buoyancyList.size());
		
		int factor = 1;
		if(lcg>lcb){
			factor = -1;
		}
		
		for(int i = 0; i < buoyancyList.size();i++){
			int currentFactor;
			float currentFactor2;
			if(buoyancyList.get(i).getCG() < lcf){
				currentFactor = factor;
				if(fwd == true){
					currentFactor2 = 0f;
				} else {
					currentFactor2 = currentFactor;
				}
			} else {
				currentFactor = factor*-1;
				if(fwd == true){
					currentFactor2 = currentFactor;
				} else {
					currentFactor2 = 0f;
				}
			}

			if (i == smallSectionIndex){
				currentFactor2 = currentFactor * 0.5f;
			}
			
			if (i == lcfSection){
				currentFactor = 0;
				currentFactor2 = 0f;
			}
			factorList.add(currentFactor);
			factorList2.add(currentFactor2);
		}
		
		float totalmoment2 = 0;
		
		for(int i = 0; i < factorList.size(); i++){
			Section x = buoyancyList.get(i);
			float xmom = factorList.get(i) * x.getCG() * x.getFactor() * x.getLength();
			float ymom = (dist2/dist1)*factorList2.get(i) * x.getCG() * x.getFactor() * x.getLength();
			totalmoment2 += (xmom + ymom);
		}
		
		float x = ((lcg*displacement)-(p*totalmoment))/ totalmoment2;
		float z = x * (dist2/dist1);
		
		float aftload = 0f;
		float fwdload;
		float lcfAft = 0f;
		float lcfFwd = 0f;
		for (int i = 0; i < factorList.size(); i++){
			Section y = buoyancyList.get(i);
			if(y.getFactor() == 1){
				lcfAft = aftload;
				break;
			}
			float midload = y.getFactor() * (p +(factorList.get(i)*x) + (factorList2.get(i)*z));
			fwdload = (2*midload)-aftload;
			y.setAftload(aftload);
			y.setFwdload(fwdload);
			aftload = fwdload;
		}
		
		fwdload = 0f;
		for (int i = (factorList.size()-1); i > -1; i--){
			Section y = buoyancyList.get(i);
			if(y.getFactor() == 1f){
				lcfFwd = fwdload;
				break;
			}
			float midload = y.getFactor() * (p +(factorList.get(i)*x)+ (factorList2.get(i)*z));
			aftload = (2*midload)-fwdload;
			y.setAftload(aftload);
			y.setFwdload(fwdload);
			fwdload = aftload;
		}
		
		Section lcfSec = buoyancyList.get(lcfSection);
		lcfSec.setAftload(lcfAft);
		lcfSec.setFwdload(lcfFwd);
		
		Section smallSec = buoyancyList.get(smallSectionIndex);
		if(fwd == true){
			lcfSec.setFwdbound(smallSec.getFwdBound());
		} else {
			lcfSec.setAftbound(smallSec.getAftBound());
		}
		
		buoyancyList.remove(smallSectionIndex);
	}
	
	//SAX document handler to handle the callback.
	private class BuoyancyHandler extends DefaultHandler{
		ArrayList<Section> data;
		
		public BuoyancyHandler(ArrayList<Section> data){
			this.data = data;
		}

		public void startElement(String uri, String name, String qname,
				Attributes att) throws SAXException {
			if(qname.equals("section")){
				Section newSect = new Section(
						new Float(att.getValue("aftbound")),
						new Float(att.getValue("fwdbound")),
						new Float(att.getValue("factor")));
				data.add(newSect);
			}
		}
	}
	
	//SAX document handler to handle the callback.
	private class LightShipHandler extends DefaultHandler{
		ArrayList<Section> data;
		
		public LightShipHandler(ArrayList<Section> data){
			this.data = data;
		}

		public void startElement(String uri, String name, String qname,
				Attributes att) throws SAXException {
			if(qname.equals("section")){
				Section newSection = new Section(
						new Float(att.getValue("aftbound")),
						new Float(att.getValue("fwdbound")),
						new Float(att.getValue("aftload")),
						new Float(att.getValue("fwdload")));
				data.add(newSection);
			}
		}
	}
	
	// Private class used to store information about the ship sections
	private class Section{
		private float aftbound;
		private float fwdbound;
		private float aftload;
		private float fwdload;
		private float factor;
		
		public Section(float aftbound, float fwdbound, float aftload, float fwdload){
			this.aftbound = aftbound;
			this.fwdbound = fwdbound;
			this.aftload = aftload;
			this.fwdload = fwdload;
		}
		
		public Section(float aftbound, float fwdbound, float factor){
			this.aftbound = aftbound;
			this.fwdbound = fwdbound;
			this.factor = factor;
		}

		public float getAftBound() {
			return aftbound;
		}

		public float getFwdBound() {
			return fwdbound;
		}

		public float getAftLoad() {
			return aftload;
		}

		public float getFwdLoad() {
			return fwdload;
		}
		
		public float getFactor() {
			return factor;
		}
		
		public void setAftbound(float aftbound) {
			this.aftbound = aftbound;
		}

		public void setFwdbound(float fwdbound) {
			this.fwdbound = fwdbound;
		}

		public float getLength(){
			return (fwdbound - aftbound);
		}
		
		public float getCG(){
			return ((fwdbound + aftbound)/2);
		}
		
		public void setAftload(float aftload) {
			this.aftload = aftload;
		}

		public void setFwdload(float fwdload) {
			this.fwdload = fwdload;
		}
	}
	
	// Method used to interpolate between value
	// refnum is the number for which we want the value
	// refnum is between refsmall and reflarge which correspond to solsmall and
	// solllargo
	private float interpolate(float refnum, float refsmall,
			float reflarge, float solsmall, float sollarge) {
		float x = refnum - refsmall;
		float y = reflarge - refsmall;
		float z = sollarge - solsmall;
		float solution = (x / y) * z;
		return (solution + solsmall);
	}
}
