//: /be/optiloading/tank/Tank.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2007 Tim Pintens. All rights reserved.
 */

package be.optiloading.tank;

import java.util.ArrayList;

import javax.swing.JOptionPane;

import be.optiloading.cargo.Cargo;
import be.optiloading.settings.Settings;

/**
 * Tank class used to stow information about the tank and calculate its center
 * of gravity.
 * 
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank"
 *         href="http://www.optiloading.be">www.optiloading.be</a>
 */

public class Tank{
	private String tankid;
	private String shortid;
	private float maxvolume;
	private float defaultdensity;
	private float aftbound;
	private float fwdbound;
	private float maxfactor;
	
	private float ullage;
	private float density;
	private float weight;
	private float volume;
	private float lcg;
	private float vcg;
	private float tcg;
	private float fsm;
	
	private String cargoid;
	
	private ArrayList<TankRecord> data;

	/**
	 * Constructor used to create a new tank.
	 * 
	 * @param tankid
	 *            String to identify the tank.
	 * @param maxvolume
	 *            Floating point value indicating the maximum value.
	 * @param defaultdensity
	 *            Floating point value indicating the default apparent density.
	 * @param aftbound
	 *            Floating point value indicating the aft bound (measured from aft).
	 * @param fwdbound
	 *            Floating point value indicating the fwd bound (measured from aft).                      
	 */
	public Tank(String tankid, String shortid, float maxvolume, float defaultdensity, 
			float aftbound, float fwdbound, ArrayList<TankRecord> data) {
		this.tankid = tankid;
		this.shortid = shortid;
		this.maxvolume = maxvolume;
		this.defaultdensity = defaultdensity;
		this.density = defaultdensity;
		this.aftbound = aftbound;
		this.fwdbound = fwdbound;
		this.data = data;
		setVolume(0f);
	}
	
	/**
	 * Method to set the volume
	 * 
	 * @param volume Float value indicating the new volume.
	 */
	public void setVolume(float volume){
		this.volume = volume;
		weight = volume * density;
		calculateDependentValues();
	}
	
	/**
	 * Method to set the volume
	 * 
	 * @param volume Float value indicating the new volume.
	 * @param density Float value indicating the apparent density of the volume.
	 */
	public void setVolume(float volume, float density){
		this.volume = volume;
		this.density = density;
		weight = volume * density;
		calculateDependentValues();
	}
	
	/**
	 * Method to set the density
	 * 
	 * @param density Float value indicating the apparent density of the volume.
	 */
	public void setDensity(float density){
		this.density = density;
		weight = volume * density;
	}
	
	/**
	 * Method to set the weight.
	 * 
	 * @param weight value indicating the new weight value.
	 */
	public void setWeight(float weight){
		this.weight = weight;
		volume = weight / density;
		calculateDependentValues();
	}
	
	/**
	 * Method to set the ullage.
	 * 
	 * @param ullage value indicating the new ullage value.
	 */
	public void setUllage(float ullage){
		this.ullage = ullage;
		try{
		for (int i = 0; i < data.size(); ++i) {
			TankRecord current = data.get(i);

			// If the volume of the object is greater then we've gone past the
			// desired volume
			if (ullage > current.getUllage()) {
				// Get the previous object to interpolate
				TankRecord previous = data.get(i - 1);
				volume = interpolate(ullage, previous.getUllage(), current.getUllage(), 
						previous.getVolume(), current.getVolume());
				lcg = interpolate(ullage, previous.getUllage(), current.getUllage(), 
						previous.getLCG(), current.getLCG());
				vcg = interpolate(ullage, previous.getUllage(), current.getUllage(), 
						previous.getVCG(), current.getVCG());
				tcg = interpolate(ullage, previous.getUllage(), current.getUllage(), 
						previous.getTCG(), current.getTCG());
				fsm = interpolate(ullage, previous.getUllage(), current.getUllage(), 
						previous.getFSM(), current.getFSM());
				break;

				// The desired volume is equal to the objects volume so we found
				// the object
			} else if (ullage == current.getUllage()) {
				volume = current.getVolume();
				lcg = current.getLCG();
				vcg = current.getVCG();
				tcg = current.getTCG();
				fsm = current.getFSM();
				break;
			}
		}
		} catch(Exception e){
			JOptionPane.showMessageDialog(null,Settings.getInstance()
					.getResourceBundle().getString("messageTankLoad"));
		}
		weight = volume * density;
	}
	
	/**
	 * Method to input a Cargo object into a tank.
	 */
	public void setCargo(Cargo cargo){
		volume = cargo.getVolume();
		density = cargo.getDensity();
		weight = cargo.getWeight();
		cargoid = cargo.getCargoid();
		calculateDependentValues();
	}
	
	// Private method to find the corresponding values for a certain volume
	private void calculateDependentValues(){
		try{
		for (int i = 0; i < data.size(); ++i) {
			TankRecord current = data.get(i);

			// If the volume of the object is greater then we've gone past the
			// desired volume
			if (volume < current.getVolume()) {
				// Get the previous object to interpolate
				TankRecord previous = data.get(i - 1);
				ullage = interpolate(volume, previous.getVolume(), current.getVolume(), 
						previous.getUllage(), current.getUllage());
				lcg = interpolate(volume, previous.getVolume(), current.getVolume(), 
						previous.getLCG(), current.getLCG());
				vcg = interpolate(volume, previous.getVolume(), current.getVolume(), 
						previous.getVCG(), current.getVCG());
				tcg = interpolate(volume, previous.getVolume(), current.getVolume(), 
						previous.getTCG(), current.getTCG());
				fsm = interpolate(volume, previous.getVolume(), current.getVolume(), 
						previous.getFSM(), current.getFSM());
				break;

				// The desired volume is equal to the objects volume so we found
				// the object
			} else if (volume == current.getVolume()) {
				ullage = current.getUllage();
				lcg = current.getLCG();
				vcg = current.getVCG();
				tcg = current.getTCG();
				fsm = current.getFSM();
				break;
			}
		}
		} catch(Exception e){
			JOptionPane.showMessageDialog(null,Settings.getInstance()
					.getResourceBundle().getString("messageTankLoad"));
		}
	}
	
	/**
	 * Method to get the tank's identification
	 * 
	 * @return String value containing the tank's identification
	 */
	public String getTankid() {
		return tankid;
	}

	/**
	 * Method to get the maximum volume of the tank
	 * 
	 * @return Floating point value containing the maximum volume of the tank
	 */
	public float getMaxvolume() {
		return (maxfactor * maxvolume);
	}
	
	/**
	 * Method to get the default density of the tank
	 * 
	 * @return Floating point value containing the default density the tank
	 */
	public float getDefaultdensity() {
		return defaultdensity;
	}
	
	/**
	 * Method to the aft bound of the tank
	 * 
	 * @return Floating point value containing the aft bound the tank
	 */
	public float getAftbound() {
		return aftbound;
	}
	
	/**
	 * Method to the fwd bound of the tank
	 * 
	 * @return Floating point value containing the fwd bound the tank
	 */
	public float getFwdbound() {
		return fwdbound;
	}
	
	/**
	 * Method to get the tank's ullage
	 * 
	 * @return Float value containing the tank's ullage
	 */
	public float getUllage() {
		return ullage;
	}
	
	/**
	 * Method to get the apparent density of the cargo in the tank
	 * 
	 * @return Floating point value containing the apparent density of the cargo
	 *         in the tank
	 */
	public float getDensity() {
		return density;
	}
	
	/**
	 * Method to get the weight of the cargo in the tank
	 * 
	 * @return Floating point value containing the weight of the cargo in the
	 *         tank
	 */
	public float getWeight() {
		return weight;
	}

	/**
	 * Method to get the volume of the cargo in the tank
	 * 
	 * @return Floating point value containing the volume of the cargo in the
	 *         tank
	 */
	public float getVolume() {
		return volume;
	}
	
	/**
	 * Method to get the Longitudinal Center of Gravity (LCG) of the tank
	 * 
	 * @return Floating point value containing the LCG of the tank
	 */
	public float getLCG() {
		return lcg;
	}

	/**
	 * Method to get the VCG of the tank
	 * 
	 * @return Floating point value containing the KG of the tank
	 */
	public float getVCG() {
		return vcg;
	}
	
	/**
	 * Method to get the VCG of the tank
	 * 
	 * @return Floating point value containing the KG of the tank
	 */
	public float getTCG() {
		return tcg;
	}

	/**
	 * Method to get the Longitudinal Free Surface Moment (FSM) of the tank
	 * 
	 * @return Floating point value containing the FSM of the tank
	 */
	public float getFSM() {
		return (fsm * getDensity());
	}
	
	/**
	 * Method to get the percantage the tank is full.
	 * 
	 * @return Floating point value containing the % full of the tank
	 */
	public float getPercentagefull() {
		return (volume/maxvolume);
	}
	
	/**
	 * Method to get the cargo id if any.
	 * 
	 * @return String value containing the cargo identification
	 */
	public String getCargoid() {
		return cargoid;
	}
	
	/**
	 * Method to get the short tank id
	 * 
	 * @return String value containing the short tank identification
	 */
	public String getShortid() {
		return shortid;
	}
	
	/**
	 * Method to get the maximum tank ullage.
	 * 
	 * @return Floating point value containing the max ullage of the tank.
	 */
	public float getMaxUllage() {
		return (data.get(0).getUllage());
	}
	
	/**
	 * Method to get the LCG value for a full tank.
	 * 
	 * @return Floating point containing the LCG value for a full tank.
	 */
	public float getLCGMax(){
		return (data.get(data.size()-1).getLCG());
	}
	
	/**
	 * Method to get the TCG value for a full tank.
	 * 
	 * @return Floating point containing the TCG value for a full tank.
	 */
	public float getTCGMax(){
		return (data.get(data.size()-1).getTCG());
	}
	
	public void setCargoid(String s) {
		this.cargoid = s;
	}
	
	/**
	 * Get the LCG value when the indicated volume is added.
	 * @param add Volume to add.
	 * @return LCG value.
	 */
	public float getLCG(float add){
		volume += (add);
		calculateDependentValues();
		float returnvalue = lcg;
		volume -= add;
		calculateDependentValues();
		return returnvalue;
	}
	
	/**
	 * Get the TCG value when the indicated volume is added.
	 * @param add Volume to add.
	 * @return TCG value.
	 */
	public float getTCG(float add){
		volume += (add);
		calculateDependentValues();
		float returnvalue = tcg;
		volume -= add;
		calculateDependentValues();
		return returnvalue;
	}
	
	/**
	 * Get the maximum volume factor.
	 * @return Floating point value with the maximum volume factor.
	 */
	public float getMaxfactor() {
		return maxfactor;
	}

	/**
	 * Set the maximum volume factor.
	 * @param maxfactor Floating point value with the maximum volume factor.
	 */
	public void setMaxfactor(float maxfactor) {
		this.maxfactor = maxfactor;
	}
	
	/**
	 * Get the max volume at 100 %.
	 * @return Floating point with the max volume at 100 %.
	 */
	public float getMax(){
		return maxvolume;
	}
	
	// Method used to interpolate between value
	// refnum is the number for which we want the value
	// refnum is between refsmall and reflarge which correspond to solsmall and
	// solllargo
	private float interpolate(float refnum, float refsmall,
			float reflarge, float solsmall, float sollarge) {
		float x = refnum - refsmall;
		float y = reflarge - refsmall;
		float z = sollarge - solsmall;
		float solution = (x / y) * z;
		return (solution + solsmall);
	}
}